from distutils.core import setup
setup(
    name='pegasus',
    packages=['pegasus'],
    version='0.1.2',
    description='Yet another PEG parser generator',
    author='Josh Junon',
    author_email='josh@junon.me',
    url='https://github.com/qix-/pegasus',
    download_url='https://github.com/qix-/pegasus/tarball/0.1.2',
    keywords=['peg', 'packrat', 'pakrat', 'parser', 'generator', 'pegasus', 'visitor', 'expression', 'grammar'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Software Development :: Compilers',
        'Topic :: Software Development :: Interpreters',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Pre-processors',
        'Topic :: Text Editors :: Text Processing',
        'Topic :: Utilities'
    ],
)
